/*
 * Copyright (c) 2002 Patrick Tullmann
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this file for any
 * purpose with or without restriction is hereby granted.
 */

import java.io.PrintStream;
import java.util.HashSet;
import java.util.Iterator;


/**
 * <width> <height>
 * <lines, south to north>
 *
 * Origin (1,1) is in south west corner;
 */
public class Board
{
	int width;
	int height;

	private final HashSet bases = new HashSet();

	public static class Base
	{
		int x;
		int y;

		int lastTurnVisited;
		int penalty;

		// should have three states: unknown, has packages, has no packages
		boolean noPackages;

		Base(int x, int y)
		{
			this.x = x;
			this.y = y;
			this.noPackages = false; // dunno, actually
			this.lastTurnVisited = -1;
			this.penalty = 0;
		}

		public String toString()
		{
			return "Base(" +x+ "," +y+ ") [penalty=" +penalty+ "]";
		}
	}


	public static class Element
	{
		public static final char OPEN_CHAR  = '.';
		public static final char WALL_CHAR  = '~';
		public static final char WATER_CHAR = '#';
		public static final char BASE_CHAR  = '@';

		public static final Element OPEN = new Element(OPEN_CHAR);
		public static final Element WALL = new Element(WALL_CHAR);
		public static final Element WATER = new Element(WATER_CHAR);
		public static final Element BASE = new Element(BASE_CHAR);

		private final String rep;
		private Element(char c)
		{
			this.rep = Character.toString(c);
		}

		public String toString()
		{
			return this.rep;
		}
	}
	
	private void checkY(int y)
	{
		if (y == 0)
			throw new Error("1-based (y)");
		if (y > height)
			throw new Error("y too large (" +y+ " > " +height+ ")");
	}
	
	private void checkX(int x)
	{
		if (x == 0)		
			throw new Error("1-based (x)");

		if (x > width)		
			throw new Error("x too large (" +x+ " > " +width+ ")");
	}

	// XXX should be byte[] (nibble[] even)
	Element[] elements;

	public Board()
	{
		this.width = this.height = -1;
	}
	
	public int width()
	{
		// 1-based
		return this.width;
	}

	public int height()
	{
		// 1-based
		return this.height;
	}

	HashSet bases()
	{
		return this.bases;
	}

	public void decayPenalities()
	{
		Iterator i = bases().iterator();
		while (i.hasNext())
		{
			Base b = (Base)(i.next());
			b.penalty -= 6;
			if (b.penalty < 0)
				b.penalty = 0;
		}
	}

	// Given base is empty (has no packages at it)
	void removeBase(Base b)
	{
		b.noPackages = true;
		this.bases.remove(b);
	}

	void addBase(int x, int y)
	{
		Trace.t(Trace.BASES, "Add base at (" +x+ ", " +y+ ")");
		// x and y are 1-based
		this.bases.add(new Base(x,y));
	}

	boolean impassableIn(int x, int y, Direction d)
	{
		int nx = x + d.dx();
		int ny = y + d.dy();
		
		if ((nx < 1) || (nx > width)
		    || (ny < 1) || (ny > height))
			return true; // is impassable

		Element el = elementAt(nx, ny);
		if ((el == Element.OPEN)
		    || (el == Element.BASE))
			return false; // not impassable

		// otherwise WALL or WATER
		return true; // is impassable
	}

	void setSize(int width, int height)
	{
		if ((this.width != -1)
		    || (this.height != -1))
			throw new Error("Re-setting board size.");
		
		this.width = width;
		this.height = height;

		// zero-based
		this.elements = new Element[width * height];
	}

	void setRow(int r, String s)
	{
		// r is zero-based
		int offset = r * this.width;
		for (int i = 0; i < this.width; i++)
		{
			char c = s.charAt(i);
			switch (c)
			{
			case Element.OPEN_CHAR:
				this.elements[offset + i] = Element.OPEN;
				break;
			case Element.WATER_CHAR:
				this.elements[offset + i] = Element.WATER;
				break;
			case Element.WALL_CHAR:
				this.elements[offset + i] = Element.WALL;
				break;
			case Element.BASE_CHAR:
				this.elements[offset + i] = Element.BASE;
				this.addBase(i+1, r+1);
				break;
			default:
				throw new Error("Expecting a map char.  Got `" +c+ "'");
			}
		}
	}

	Element elementAt(int x, int y)
	{
		checkX(x);
		checkY(y);

		return elements[(y-1)*width + (x-1)];
	}

	void dump(PrintStream pw)
	{
		for (int y = height-1; y > 0; y--)
		{
			for (int x = 1; x <= width; x++)
			{
				pw.print(elementAt(x,y).toString());
			}
			pw.println();
		}
	}
}
