import java.util.Iterator;
import java.util.NoSuchElementException;


// Downloaded from http://www.sosnoski.com/Performance/Code/IntKeyedHashtable.java
// On August 31, 2002
//
// Added: toString, iterator

/**
 * Copyright(C) 1999-2000, Sosnoski Software Solutions, Inc. General permission
 * is hereby granted to reuse this software for any purpose, including use
 * in applications developed for distribution and sale, provided this
 * copyright notice and usage policy is retained intact. The software is
 * provided "AS IS", with no warranty of any kind. ALL USE IS STRICTLY AT
 * THE USER'S RISK, AND IN NO CASE WILL SOSNOSKI SOFTWARE SOLUTIONS, INC.
 * BE LIABLE FOR ANY DIRECT OR INDIRECT COSTS OR DAMAGES OF ANY FORM ARISING
 * THROUGH OR FROM THE USE OF THIS SOFTWARE. 
 */

/**
 * Hashtable using primitive <code>int</code> values as keys, with no extra 
 * objects created. This uses the standard hash technique of adding a standard 
 * offset to the computed position in the table when a collision occurs, so
 * that several checks may potentially be necessary in order to determine
 * if a particular key value is present in the table (since if the slot
 * calculated by the hash function is occupied by a different entry it's
 * still necessary to check the offset slot for a match, and then the 
 * offset slot from that, until a empty slot is found).<p>
 *
 * Each time the number of entries present in the table grows above the
 * specified fraction full the table is expanded to twice its prior size
 * plus one (to ensure the size is always an odd number). The collision
 * offset is set to half the table size rounded down, ensuring that it
 * will cycle through all slots of the table before returning to the
 * original slot.<p>
 *
 * This implementation is designed for cases where the keys are already
 * present in the hashtable entries, and requires that the entries 
 * implement the inner interface <code>IntKey</code> so that the
 * table can access the key value as necessary. The hashing technique used
 * to convert the key values to initial table slots is also naive, and may
 * perform poorly in some circumstances.<p>
 *
 * It does not include support for removing entries from the table. If
 * support for removals is needed it can easily be added, but remember that
 * removing an entry will require following its collision chain until an
 * empty slot is found, recursively moving down any entries found which hash
 * to any of the slots visited on the chain.
 * 
 * @author Dennis M. Sosnoski
 * @version 1.0
 */
public class IntKeyedHashtable
{
	/**
	 * Interface which must be implemented by the entries in the table. This 
	 * allows easy access to key values contained within the entry objects
	 * themselves.
	 */
	public interface IntKey
	{
		int keyInt();
	}

	private static class IKHIterator
	implements java.util.Iterator
	{
		final IntKey[] table;
		int current;

		private void scanToNext()
		{
			while ((current < table.length)
			       && (table[current] == null))
				current++;
		}

		IKHIterator(IntKey[] table)
		{
			this.table = table;
			this.current = 0;

			// find first non-null entry
			scanToNext();
		}
		
		public boolean hasNext()
		{
			return (current < table.length);
		}
		
		public Object next()
		{
			if (current < table.length)
			{
				if (table[current] == null)
					throw new Error("Can't happen.");
				Object rc = table[current++];
				scanToNext();
				return rc;
			}
			throw new NoSuchElementException("Iterated past the end...");
		}

		public void remove()
		{
			throw new UnsupportedOperationException("No remove on IntKeyedHashtables");
		}
	}


	/** Default fill fraction allowed before growing table. */
	protected static final double DEFAULT_FILL = 0.3d;
	
	/** Minimum size used for hash table. */
	protected static final int MINIMUM_SIZE = 31;
	
	/** Hash value multiplier to scramble bits before calculating slot. */
	protected static final int KEY_MULTIPLIER = 517;
	
	/** Fill fraction allowed for this hash table. */
	protected double fillFraction;
	
	/** Number of entries present in table. */
	protected int entryCount;
	
	/** Entries allowed before growing table. */
	protected int entryLimit;
	
	/** Offset added (modulo table size) to slot number on collision. */
	protected int hitOffset;
	
	/** Array of table slots. */
	protected IntKey[] hashTable;

	/**
	 * Constructor with full specification.
	 * 
	 * @param count number of values to assume in initial sizing of table
	 * @param fill fraction full allowed for table before growing
	 */
	
	public IntKeyedHashtable(int count, double fill) {
		
		// check the passed in fill fraction
		if (fill <= 0.0d || fill >= 1.0d) {
			throw new IllegalArgumentException("fill value out of range");
		}
		fillFraction = fill;
		
		// compute initial table size (ensuring odd)
		int size = Math.max((int) (count / fillFraction), MINIMUM_SIZE);
		size += (size + 1) % 2;
		
		// initialize the table information
		entryLimit = (int) (size * fillFraction);
		hitOffset = size / 2;
		hashTable = new IntKey[size];
	}

	/**
	 * Constructor with only size supplied. Uses default value for fill
	 * fraction.
	 * 
	 * @param count number of values to assume in initial sizing of table
	 */
	
	public IntKeyedHashtable(int count) {
		this(count, DEFAULT_FILL);
	}

	/**
	 * Default constructor.
	 */
	
	public IntKeyedHashtable() {
		this(0, DEFAULT_FILL);
	}

	/**
	 * Copy (clone) constructor.
	 * 
	 * @param from instance being copied
	 */
	
	public IntKeyedHashtable(IntKeyedHashtable from) {
		fillFraction = from.fillFraction;
		entryCount = from.entryCount;
		entryLimit = from.entryLimit;
		hitOffset = from.hitOffset;
		hashTable = new IntKey[from.hashTable.length];
		System.arraycopy(from.hashTable, 0, hashTable, 0, hashTable.length);
	}

	/**
	 * Get number of items in table.
	 * 
	 * @return item count present
	 */
	
	public int size() {
		return entryCount;
	}

	/**
	 * Find slot number for entry. Starts at the slot found by the hashed
	 * key value. If this slot is already occupied, it adds the collision
	 * offset (modulo the table size) to the slot number and checks that
	 * slot, repeating until an unused slot is found.
	 * 
	 * @param entry entry to be added to table
	 * @return slot at which entry was added
	 */
	
	protected int assignSlot(IntKey entry) {
		int offset = (entry.keyInt() * KEY_MULTIPLIER &
			Integer.MAX_VALUE) % hashTable.length;
		while (hashTable[offset] != null) {
			offset = (offset + hitOffset) % hashTable.length;
		}
		hashTable[offset] = entry;
		return offset;
	}

	/**
	 * Expand the table. Increases the table size to twice the previous size
	 * plus one, reinserting each entry from the smaller table into the new
	 * one.
	 */
	
	protected void expandTable() {
		
		// initialize for the increased table size
		int size = hashTable.length * 2 + 1;
		entryLimit = (int) (size * fillFraction);
		hitOffset = hashTable.length;
		IntKey[] old = hashTable;
		hashTable = new IntKey[size];
		
		// reinsert all entries into new table
		for (int i = 0; i < old.length; i++) {
			IntKey entry = old[i];
			if (entry != null) {
				assignSlot(entry);
			}
		}
	}

	/**
	 * Add an entry to the table. Note that nothing prevents multiple entries
	 * with the same key, but the entry returned by that key will be
	 * undetermined in this case.
	 * 
	 * @param entry entry to be added to table
	 */
	
	public void add(IntKey entry) {
		if (++entryCount > entryLimit) {
			expandTable();
		}
		assignSlot(entry);
	}

	/**
	 * Find an entry in the table.
	 * 
	 * @param key for entry to be returned
	 * @return entry for key, or <code>null</code> if none
	 */
	
	public IntKey get(int key) {
		int offset = (key * KEY_MULTIPLIER &
			Integer.MAX_VALUE) % hashTable.length;
		IntKey entry;
		while ((entry = hashTable[offset]) != null) {
			if (key == entry.keyInt()) {
				return entry;
			}
			offset = (offset + hitOffset) % hashTable.length;
		}
		return null;
	}

	public String toString()
	{
		StringBuffer sbuf = new StringBuffer(entryCount * 20);
		sbuf.append("[");
		for (int i = 0; i < hashTable.length; i++)
		{
			if (hashTable[i] != null)
				sbuf.append(" <")
					.append(i)
					.append("> -> ")
					.append(hashTable[i].toString());
		}
		sbuf.append("]");
		return sbuf.toString();
	}

	public Iterator iterator()
	{
		return new IKHIterator(this.hashTable);
	}

	/**
	 * Construct a copy of the table.
	 * 
	 * @return shallow copy of table
	 */
	
	public Object clone() {
		return new IntKeyedHashtable(this);
	}

}
