/*
 * Copyright (c) 2002 Patrick Tullmann
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this file for any
 * purpose with or without restriction is hereby granted.
 */

import java.util.Iterator;

public class Robot_BaseHoming
	extends Robot
{
	private int destX;
	private int destY;
	private final PackageSet heldPackages;
	private int turnCt = 0;

	// Subclasses for heading to base vs. delivering package logic

	public Robot_BaseHoming()
	{
		super();
		heldPackages = new PackageSet();
		destX = -1;
		destY = -1;
	}

	public void placed(int x, int y)
	{
		// I'm lost:
		this.destX = -1;
		this.destY = -1;

		super.placed(x,y);
	}
	
	public void picked(Package pkg)
	{
		this.heldPackages.add(pkg);
		super.picked(pkg);
	}

	public void dropped(Package pkg)
	{
		this.heldPackages.remove(pkg);
		super.dropped(pkg);
	}

	public void moved(Direction d)
	{
		super.moved(d);
	}

	// XXX Utils:
	static int manhattanDist(int x1, int y1, int x2, int y2)
	{
		int dx = x2 - x1;
		int dy = y2 - y1;
		if (dx < 0)
			dx = -dx;
		if (dy < 0)
			dy = -dy;
		return dx+dy;
	}

	Direction nextOnPathTo(Board board, int destX, int destY)
	{
		AStarSearch searcher = new AStarSearch(board, 
						       this.x, this.y,
						       destX, destY);
		return searcher.search();
	}

	Command deliverCommand(PackageSet.PackageList pl)
	{
		int pkgct = pl.size();
		int[] pkglist = new int[pkgct];
		Iterator iter = pl.iterator();
		int i = 0;
		while (iter.hasNext())
			pkglist[i++] = ((Package)(iter.next())).id();
		
		return Command.drop(useFunds(1), pkglist);
	}

	private Board.Base pickHomeBase(Board board, WorldState worldState)
	{
		if (board.bases().isEmpty())
			return null;

		Iterator i  = board.bases().iterator();
		Board.Base minBase = null;
		int minDist = 1000 + 1000; // Bigger than possible dist squared

		// XXX closest via Manhattan dist, not actual path dist.

		while (i.hasNext())
		{
			Board.Base b = (Board.Base)i.next();

			// If I'm standing on it, there are no packages, so its
			// really stupid
			if ((b.x == this.x)
			    && (b.y == this.y))
			{
				b.penalty += 200;
			}

			if (minBase == null)
			{
				minBase = b;
				minDist = manhattanDist(b.x, b.y, this.x, this.y) + b.penalty;
			}
			else 
			{
				int dist = manhattanDist(b.x, b.y, this.x, this.y) + b.penalty;
				if (dist < minDist)
				{
					minBase = b;
					minDist = dist;
				}
			}
		}

		return minBase;
	}

	private void pickDestination(Board board, WorldState worldState)
	{
		Board.Base b = pickHomeBase(board, worldState);
		if (b != null)
		{
			if ((b.lastTurnVisited != -1)
			    && ((this.turnCt - b.lastTurnVisited) < 3))
			{
				/*
				 * Bleh, too many visits (probably other bots)
				 * Penalize this base for later.
				 */
				b.penalty += 40;
			}

			// Skip completely if penalty on base is huge...
			if (b.penalty < 180)
			{
				msg("Picked destination: base: " +b);
				this.destX = b.x;
				this.destY = b.y;
				b.lastTurnVisited = this.turnCt;
				return;
			}
		}
		
		Package p = worldState.pickRandomKnownPackage();
		if (p != null)
		{
			msg("Picked destination: random package: " +p);
			this.destX = p.x();
			this.destY = p.y();
			return;
		}

		WorldState.OtherRobot r = worldState.pickRandomKnownRobot();
		if (r != null)
		{
			msg("Picked destination: random robot: " +r);
			this.destX = r.x;
			this.destY = r.y;
			return;
		}

		// Ah, well.
	}


	public Command command(Board board,
			       WorldState worldState,
			       Package[] currentPackages)
	{
		turnCt++;

		msg("Sitting at (" +this.x+ "," +this.y+ 
		    "), cap=" +this.remainingCapacity+ 
		    ", pkgct=" +this.heldPackages.packageCount());
		if ((x == -1)
		    || (y == -1))
			throw new Error("Where am I?");


		// If any packages belong here, try to drop them
		PackageSet.PackageList pl = heldPackages.getPackageList(this.x, this.y);
		if ((pl != null) 
		    && (! pl.isEmpty()))
		{
			msg("Delivering a pile");
			return deliverCommand(pl);
		}

		// If there are any available packages here, try to pick them up
		if (currentPackages.length != 0)
		{
			msg("Mmmm! " +currentPackages.length+ " Packages!");
			int i = 0;
			int availableStorage = this.remainingCapacity;
			int[] picks = new int[currentPackages.length];
			int picked = 0;

			for (i = 0; i < currentPackages.length; i++)
			{
				Package p = currentPackages[i];
				int weight = p.weight();
				if (availableStorage > weight)
				{
					picks[picked++] = p.id();
					availableStorage -= weight;
				}
			}

			if (picked > 0)
			{
				msg("Trying for " +picked+ " of them.");

				// Toss any plan I may have had
				this.destX = -1;
				this.destY = -1; // xxx redundant

				return Command.pick(useFunds(1), picks, picked);
			}
			

			msg("No useful packages!");
			// Fall through
		}
		
		// If we're currently at our destintation, we've
		// completed the current plan.
		if ((this.x == this.destX)
		    && (this.y == this.destY))
		{
			this.destX = -1;
			this.destY = -1; // xxx redundant
		}

		// If we have packages and haven't got a destination, pick a destination.
		// XXX should go fetch more packages if they're close and I've got lots
		// of room... ?
		if ((! heldPackages.isEmpty())
		    && (destX == -1))
		{
			Package bestDest = heldPackages.bestToDeliver();
			if (bestDest != null)
			{
				this.destX = bestDest.destX();
				this.destY = bestDest.destY();
				//msg("Heading towards (" +this.destX+ ", " +this.destY+ ")");
			}
			else
			{
				throw new Error("Umm... have packages, but no destination for them?!!");
			}
		}

		// Otherwise if no plan (and no packages), pick a base to go to...
		if (this.destX == -1)
		{
			pickDestination(board, worldState);
			if (this.destX == -1)
			{
				msg("Nothing to do... twiddling thumbs...");
				return Command.noop(useFunds(1));
			}
		}
		
			
		// If still without plan, am greatly hosed
		if ((this.destX == -1)
		    || (this.destY == -1))
			throw new Error("Bogus destination (" +this.destX+ ", " +this.destY+ ")");

		Direction d = this.nextOnPathTo(board, this.destX, this.destY);

		msg("Will try to move " +d+ " (towards " +this.destX+ ", " +this.destY+ ")");
		return Command.move(useFunds(1), d);
	}
}
