/*
 * Copyright (c) 2002 Patrick Tullmann
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this file for any
 * purpose with or without restriction is hereby granted.
 */

import java.util.StringTokenizer;

/**
 * Use by setting property "trace" to a string of ",| " separated
 * string constants from the table below.
 */
/*package*/ class Trace
{
  private static class Flag
  {
    final String name;
    final long mask;
    
    static final Flag[] flags = new Flag[64];
    static int fi = 0;

    Flag(String name, long mask)
    {
      this.name = name;
      this.mask = mask;

      synchronized(flags)
      {
	flags[fi++] = this;
      }
    }

    static Flag find(String name)
    {
      synchronized(flags)
      {
	for (int i = 0; i < flags.length; i++)
	{
	  if ((flags[i] != null)
	      && (flags[i].name.equals(name)))
	    return flags[i];
	}
	throw new Error("No trace flag with name " +name);
      }
    }
  }
  
  static final Flag NONE        = new Flag("NONE",        0x00000000);
  static final Flag SEND_CMD    = new Flag("SEND_CMD",    0x00000002);
  static final Flag SEND_MSG    = new Flag("SEND_MSG",    0x00000004);
  static final Flag BASES       = new Flag("BASES",       0x00000008);
  static final Flag UPDATEWORLD = new Flag("UPDATEWORLD", 0x00000010);
  static final Flag PKGS        = new Flag("PKGS",        0x00000020);
  static final Flag TOKER       = new Flag("TOKER",       0x00000040);
  static final Flag SERVER      = new Flag("SERVER",      0x00000100|Trace.UPDATEWORLD.mask);
  static final Flag PACKAGESET  = new Flag("PACKAGESET",  0x00000200);
  static final Flag ASTAR       = new Flag("ASTAR",       0x00000400);
  static final Flag ALL         = new Flag("ALL",         0xFFFFFFFF);

  private static long currentTraceMask;
  private static final String traceInitProp = "trace";

  static
  {
    String propVal = System.getProperty(traceInitProp, "NONE");
    StringTokenizer propWalk = new StringTokenizer(propVal, ",| ");
    currentTraceMask |= 0;
    while(propWalk.hasMoreTokens())
    {
      String flag = propWalk.nextToken();
      currentTraceMask |= Flag.find(flag).mask;
    }
  }

  static boolean flagSet(Flag f)
  {
    return (f.mask & currentTraceMask) != 0;
  }

  static void write(String s)
  {
    synchronized(Trace.class)
    {
      System.err.println(s);
    }
  }

  static void write(String pfx, String s)
  {
    synchronized(Trace.class)
    {
      System.err.print(pfx);
      System.err.print(": ");
      System.err.println(s);
    }
  }

  static void write(Throwable t)
  {
    synchronized(Trace.class)
    {
      t.printStackTrace(System.err);
    }
  }

  static void write(String pfx, Throwable t)
  {
    synchronized(Trace.class)
    {
      System.err.print(pfx);
      System.err.print(": ");
      t.printStackTrace(System.err);
    }
  }

  static void t(Flag f, String s)
  {
    if (flagSet(f))
      write(f.name, s);
  }

  static void t(Flag f, Throwable t)
  {
    if (flagSet(f))
      write(f.name, t);
  }
}
